import hashlib
from multiprocessing import Event, Queue
from typing import Callable, Dict, Optional

import bech32
import secp256k1


_BECH32_CHARSET = "023456789acdefghjklmnpqrstuvwxyzACDEFGHJKLMNPQRSTUVWXYZ"


def generate_wallet():
    priv_key: secp256k1.PrivateKey = secp256k1.PrivateKey()
    pub_key: secp256k1.PublicKey = priv_key.pubkey
    byte_arr = pub_key.serialize(compressed=True)
    s = hashlib.new("sha256", byte_arr).digest()
    r = hashlib.new("ripemd160", s).digest()
    bech_addr = bech32.bech32_encode("cosmos", bech32.convertbits(r, 8, 5))
    return bech_addr, byte_arr.hex(), priv_key.serialize()


def ends_with(suffix, bech_addr: str) -> bool:
    return bech_addr.endswith(suffix)


def starts_with(prefix, bech_addr: str) -> bool:
    return bech_addr[7:].startswith(prefix)


def contains(vanity, bech_addr: str) -> bool:
    return vanity in bech_addr


def letters(vanity, bech_addr: str) -> bool:
    return vanity <= sum(not c.isdigit() for c in bech_addr[7:])


def digits(vanity, bech_addr: str) -> bool:
    return vanity <= sum(c.isdigit() for c in bech_addr[7:])


def _is_valid_addr(predicates, candidate: str) -> Optional[str]:
    is_valid = set()
    for predicate, vanity in predicates.items():
        is_valid.add(predicate(vanity, candidate))
    if False in is_valid:
        return None
    return candidate


def find_vanity_addr(predicates: Dict[Callable, str], event: Event(), queue: Queue):
    while not event.is_set():
        candidate, pub, priv = generate_wallet()
        if _is_valid_addr(predicates, candidate):
            queue.put((candidate, pub, priv))



import argparse
import multiprocessing as mp
from multiprocessing import Manager, Pool

from vanitycosmos.vanitycosmos import (
    _BECH32_CHARSET,
    contains,
    digits,
    ends_with,
    find_vanity_addr,
    letters,
    starts_with,
)


def main():
    parser = argparse.ArgumentParser(
        description="Cosmosvanity - Create custom cosmos addresses."
    )
    parser.add_argument(
        "--startswith",
        type=str,
        help="Find an address ending with the provided argument.",
    )
    parser.add_argument(
        "--endswith",
        type=str,
        help="Find an address starting with the provided argument.",
    )
    parser.add_argument(
        "--contains", type=str, help="Find an address containing the provided argument."
    )
    parser.add_argument(
        "--letters",
        type=int,
        help="Find an address containing the provided number of letters.",
    )
    parser.add_argument(
        "--digits",
        type=int,
        help="Find an address containing the provided number of digits.",
    )
    parser.add_argument(
        "-n", type=int, help="Number of addresses to search for.", default=1
    )
    args = parser.parse_args()

    vanity_args = {}
    if args.startswith and _is_valid_bech32(args.startswith):
        vanity_args[starts_with] = str(args.startswith)
    if args.endswith and _is_valid_bech32(args.endswith):
        vanity_args[ends_with] = str(args.endswith)
    if args.contains and _is_valid_bech32(args.contains):
        vanity_args[contains] = args.contains
    if args.letters and args.letters:
        vanity_args[letters] = args.letters
    if args.digits and args.digits:
        vanity_args[digits] = args.digits

    with Pool(processes=mp.cpu_count()) as pool:
        manager = Manager()
        event = manager.Event()
        queue = manager.Queue()
        for i in range(args.n):
            for _ in range(mp.cpu_count()):
                pool.apply_async(find_vanity_addr, args=(vanity_args, event, queue))

        vanity_addresses = []
        address = None
        while not address:
            address = queue.get()
            vanity_addresses.append(address)
            address = None
            if len(vanity_addresses) >= args.n:
                event.set()
                break
        for address in vanity_addresses:
            print(address)


def _is_valid_bech32(vanity: str):
    if not all(c in _BECH32_CHARSET for c in vanity):
        print(f"Not valid bech32 vanity input for {vanity}")
        exit(1)
    return True


if __name__ == "__main__":
    main()



import time

from unittest.mock import patch
from multiprocessing import Event, Queue, Process
from vanitycosmos.vanitycosmos import (
    generate_wallet,
    starts_with,
    ends_with,
    letters,
    digits,
    find_vanity_addr,
)

_WALLET = (
    "cosmos16zfzju9q6y9nfqeaxp3yvfvzvrx9rlam4ezznm",
    "0374a436044b4904bbd7a074b098d65fad39fc5b66f28da8440f10dbcf86568429",
    "b66022fff8b6322f8b8fa444d6d097457b6b9e7bb05add5b75f9c827df7bd3b6",
)


def test_generate_wallet():
    mocked_rand = b"\xb6`\"\xff\xf8\xb62/\x8b\x8f\xa4D\xd6\xd0\x97E{k\x9e{\xb0Z\xdd[u\xf9\xc8'\xdf{\xd3\xb6"

    with patch("os.urandom", return_value=mocked_rand):
        wallet = generate_wallet()
        assert wallet == _WALLET


def test_starts_with():
    assert starts_with("aaa", "cosmos1aaa")


def test_endswith():
    assert ends_with("aaa", "cosmos1ldsfjalsaaa")


def test_letters():
    assert letters(5, "cosmos1aaaaa6")


def test_digits():
    assert digits(5, "cosmos112345")


def _trigger_event(event: Event):
    time.sleep(1)
    event.set()


def test_find_vanity_address_success():
    predicates = {
        starts_with: "0000",
        ends_with: "0000",
    }
    event = Event()
    queue_mock = Queue()
    process = Process(target=_trigger_event, args=(event,))
    process.start()
    with patch("vanitycosmos.vanitycosmos.starts_with", return_value=True), patch(
        "vanitycosmos.vanitycosmos.ends_with", return_value=True
    ):
        find_vanity_addr(predicates, event, queue_mock)


def test_find_vanity_address_fails():
    predicates = {
        starts_with: "0000",
        ends_with: "0000",
    }
    event = Event()
    queue_mock = Queue()
    process = Process(target=_trigger_event, args=(event,))
    process.start()
    with patch("vanitycosmos.vanitycosmos.starts_with", return_value=False), patch(
        "vanitycosmos.vanitycosmos.ends_with", return_value=False
    ):
        find_vanity_addr(predicates, event, queue_mock)



from os import path

import setuptools

INSTALL_REQUIRES = ["bech32", "secp256k1"]

TEST_REQUIRES = ["pytest", "pytest-cov", "pytest-mock"]


def read(file_name: str) -> str:
    """Helper to read README."""
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, file_name), encoding="utf-8") as f:
        return f.read()


setuptools.setup(
    name="vanity-cosmos",
    version="0.1.0",
    author="Paul Angerer",
    author_email="etimoz@users.noreply.github.com",
    description="Tool for generating custom cosmos addresses.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/etimoz/vanity-cosmos/",
    keywords="cosmos blockchain atom cryptocurrency address generator vanity",
    zip_safe=False,
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    extras_require={"test": TEST_REQUIRES + INSTALL_REQUIRES,},
    python_requires=">=3.6",
    install_requires=INSTALL_REQUIRES,
    entry_points={"console_scripts": ["vanitycosmos = vanitycosmos.__main__:main"]},
)
